"""
Adapted from https://github.com/milesial/Pytorch-UNet
"""


import torch
import torch.nn as nn
import torch.nn.functional as F



class DoubleConv(nn.Module):

    """ (conv2d + batchnorm + relu) * 2 """

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if mid_channels == None:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=3, padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)



class MaxpoolConv(nn.Module):

    """ (maxpool + doubleconv) * 1 """

    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(kernel_size=2),
            DoubleConv(in_channels, out_channels)
        )

    def forward(self, x):
        return self.maxpool_conv(x)



class UpConv(nn.Module):

    """ (transposeConv/upsample + doubleconv) * 1 """

    def __init__(self, in_channels, out_channels, mode='bilinear'):
        super().__init__()
        if mode == 'bilinear':
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)
        if mode == 'transpose':
            self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels)

    def forward(self, x, feature):
        x = self.up(x)
        diff_y = feature.size()[2] - x.size()[2] 
        diff_x = feature.size()[3] - x.size()[3]
        x = F.pad(x, [diff_x // 2, diff_x - diff_x // 2,
                      diff_y // 2, diff_y - diff_y // 2])
        x = torch.cat([feature, x], dim = 1)
        return self.conv(x)



class OutConv(nn.Module):

    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)

    def forward(self, x):
        return self.conv(x)



class UNet(nn.Module):
    def __init__(self, n_channels, n_classes, up_mode='bilinear'):
        super(UNet, self).__init__()
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.up_mode = up_mode

        self.input = DoubleConv(n_channels, 64)
        self.down1 = MaxpoolConv(64, 128)
        self.down2 = MaxpoolConv(128, 256)
        self.down3 = MaxpoolConv(256, 512)
        factor = 2 if up_mode == 'bilinear' else 1 
        self.down4 = MaxpoolConv(512, 1024 // factor)

        self.up1 = UpConv(1024, 512 // factor, up_mode)
        self.up2 = UpConv(512, 256 // factor, up_mode)
        self.up3 = UpConv(256, 128 // factor, up_mode)
        self.up4 = UpConv(128, 64, up_mode)
        self.out = OutConv(64, n_classes) 

    def forward(self, x):
        x1 = self.input(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)

        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)

        logits = self.out(x)
        return logits
