import torch
from torch.utils.data import Dataset
from torch.utils.data import Subset
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader
import numpy as np
import skimage.io as io
import os
import IPython
import matplotlib.pyplot as plt
import torchvision.transforms as transforms


# split to train, val, test sets
def split_dataset(dataset, val_split=0.2, test_split=0.1):
    index_list = list(range(len(dataset)))
    train_index, test_index = train_test_split(index_list,
                                test_size=test_split, random_state=1)
    train_index, val_index = train_test_split(train_index, test_size=val_split,
                                              random_state=1)
    datasets = {}
    datasets['train'] = Subset(dataset, train_index)
    datasets['val'] = Subset(dataset, val_index)
    datasets['test'] = Subset(dataset, test_index)
    return datasets


class HDRDataset(Dataset):

    def __init__(self, root):
        self.root_path = root
        self.ldr_neg_path = os.path.join(root, 'LDR_exposure_-2')
        self.ldr_pos_path = os.path.join(root, 'LDR_exposure_+2')
        self.ldr_zero_path = os.path.join(root, 'LDR_exposure_0')
        self.hdr_path = os.path.join(root, 'HDR')
        # set the size of the input tensor
        self.transform = transforms.Compose([transforms.ToTensor(),
                                             transforms.Resize(size=(256,256), interpolation=transforms.InterpolationMode.NEAREST)])

        self.init_image_paths()

    def init_image_paths(self):
        self.neg_imgs = sorted(os.listdir(self.ldr_neg_path))
        self.neg_imgs = [os.path.join(self.ldr_neg_path, i) for i in self.neg_imgs]
        self.pos_imgs = sorted(os.listdir(self.ldr_pos_path))
        self.pos_imgs = [os.path.join(self.ldr_pos_path, i) for i in self.pos_imgs]
        self.zero_imgs = sorted(os.listdir(self.ldr_zero_path))
        self.zero_imgs = [os.path.join(self.ldr_zero_path, i) for i in self.zero_imgs]
        self.hdr_imgs = sorted(os.listdir(self.hdr_path))[1:]
        self.hdr_imgs = [os.path.join(self.hdr_path, i) for i in self.hdr_imgs]
        self.total_num_imgs = len(self.neg_imgs)
        return

    def __len__(self):
        return self.total_num_imgs

    def __getitem__(self, index):
        neg_img = io.imread(self.neg_imgs[index]).astype(np.float32)/255.0
        pos_img = io.imread(self.pos_imgs[index]).astype(np.float32)/255.0
        zero_img = io.imread(self.zero_imgs[index]).astype(np.float32)/255.0
        hdr_img = io.imread(self.hdr_imgs[index]).astype(np.float32)
        ldr_img = np.concatenate((neg_img, zero_img, pos_img), axis=2)
        ldr_img, hdr_img = self.transform(ldr_img), self.transform(hdr_img)
        return ldr_img, hdr_img


if __name__ == '__main__':

    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')

    batchsize=2
    data_root_path = '/home/cz_newacc/EE367/367_HDR/data'
    dataset = HDRDataset(data_root_path)

    all_datasets = split_dataset(dataset)
    dataloaders = {}
    dataloaders['train'] = DataLoader(all_datasets['train'], batchsize, shuffle=True)
    dataloaders['val'] = DataLoader(all_datasets['val'], batchsize, shuffle=True)
    dataloaders['test'] = DataLoader(all_datasets['test'], batchsize, shuffle=True)

    print(dataloaders['train'])

    ldr, hdr = next(iter(dataloaders['train']))
    print(ldr.shape, hdr.shape)
    
    # check LDR images
    plt.imshow(ldr[0,0:3,:,:].numpy().transpose(1,2,0))
    plt.savefig('/home/cz_newacc/EE367/367_HDR/test_dataloader_output/dr1.png')
    plt.imshow(ldr[1,0:3,:,:].numpy().transpose(1,2,0))
    plt.savefig('/home/cz_newacc/EE367/367_HDR/test_dataloader_output/dr2.png')
    plt.imshow(ldr[0,3:6,:,:].numpy().transpose(1,2,0))
    plt.savefig('/home/cz_newacc/EE367/367_HDR/test_dataloader_output/dr3.png')
    plt.imshow(ldr[1,3:6,:,:].numpy().transpose(1,2,0))
    plt.savefig('/home/cz_newacc/EE367/367_HDR/test_dataloader_output/dr4.png')
    plt.imshow(ldr[0,6:9,:,:].numpy().transpose(1,2,0))
    plt.savefig('/home/cz_newacc/EE367/367_HDR/test_dataloader_output/dr5.png')
    plt.imshow(ldr[1,6:9,:,:].numpy().transpose(1,2,0))
    plt.savefig('/home/cz_newacc/EE367/367_HDR/test_dataloader_output/dr6.png')
    
    # check HDR images
    # print(hdr)
    # cv2.imwrite("hdr1.hdr", hdr[0,:,:,:].numpy().transpose(2,1,0))
    # cv2.imwrite("hdr2.hdr", hdr[1,:,:,:].numpy().transpose(2,1,0))
