import torch
import torch.optim as optim
import torch.nn as nn
from UNet import UNet
from dataloader import HDRDataset, split_dataset
from torch.utils.data import DataLoader
from tqdm import tqdm
import numpy as np
from pypher.pypher import psf2otf


def tv_regularizer(img, output, label, lam=1):
    loss = 0
    for i in range(img.shape[0]):
        for j in range(3):
            dx = np.array([[-1., 1.]])
            dy = np.array([[-1.], [1.]])
            dxFT = torch.from_numpy(psf2otf(dx, (256,256))).to(device)
            dyFT = torch.from_numpy(psf2otf(dy, (256,256))).to(device)
            dxyFT = torch.stack((dxFT, dyFT), axis=0)
            grad_fn = lambda x: torch.real(torch.fft.ifft2(torch.stack((torch.fft.fft2(x),torch.fft.fft2(x)), axis=0) * dxyFT))
            grad = grad_fn(output[i,j,:,:])
            loss += torch.sum(torch.abs(grad[0]) + torch.abs(grad[1]))
    return loss

def cosine_similarity(x, gt, lam):
    cosine_sim = 0.0
    for i in range(x.shape[0]):
        this_x, this_gt = x[i,:,:,:], gt[i,:,:,:]
        nom = torch.sum(this_x * this_gt)
        denom = torch.norm(this_x) * torch.norm(this_gt)
        cosine_sim += nom/denom
    return lam * (1 - cosine_sim / x.shape[0])
    

def calc_psnr(x, gt):
    out = 10 * np.log10(1 / ((x - gt)**2).mean().item())
    return out


def prepare_dataloader(root_path, batchsize):
    data_root_path = root_path
    dataset = HDRDataset(data_root_path)
    all_datasets = split_dataset(dataset)
    dataloaders = {}
    dataloaders['train'] = DataLoader(all_datasets['train'], batchsize, shuffle=True)
    dataloaders['val'] = DataLoader(all_datasets['val'], batchsize, shuffle=True)
    dataloaders['test'] = DataLoader(all_datasets['test'], batchsize, shuffle=True)
    return dataloaders

# training for each epoch
def train_step(epoch, train_dataloader, model, optimizer, criterion):
    train_loss = []
    train_psnr = []
    lam = 0.001
    model.train()
    pbar = tqdm(train_dataloader)
    pbar.set_description(f"Train Epoch {epoch+1}")
    for i, data in enumerate(pbar):
        img, label = data
        img = img.to(device)
        label = label.to(device)
        optimizer.zero_grad()
        output = model(img)
        psnr = calc_psnr(output.detach().cpu().numpy(), label.detach().cpu().numpy())
        # loss = criterion(output, label) + cosine_similarity(output, label, lam)
        loss = criterion(output, label) + tv_regularizer(img, output, label, lam)
        train_loss.append(loss.detach().cpu().numpy())
        train_psnr.append(psnr)
        loss.backward()
        optimizer.step()
        if i % 10 == 0:
            print(f'!!! Train loss = {loss}; Train psnr = {psnr} !!!')
    pbar.close()
    return sum(train_loss)/len(train_loss), sum(train_psnr)/len(train_psnr), model

# validation after each epoch
def validation_step(epoch, val_dataloader, model, criterion):
    val_loss = []
    val_psnr = []
    lam = 0.001

    model.eval()
    with torch.no_grad():
        pbar = tqdm(val_dataloader)
        pbar.set_description(f'Val Epoch {epoch+1}')
        for i, data in enumerate(pbar):
            img, label = data
            img = img.to(device)
            label = label.to(device)
            output = model(img)
            # loss = criterion(output, label) + cosine_similarity(output, label, lam)
            loss = criterion(output, label) + tv_regularizer(output, label, lam)
            psnr = calc_psnr(output.cpu().numpy(), label.cpu().numpy())
            val_loss.append(loss.cpu().numpy())
            val_psnr.append(psnr)
            if i % 10 == 0:
                print(f'Validation loss = {loss}; Validation psnr = {psnr}')
    # file_name = f'/home/cz_newacc/EE367/367_HDR/val_output_img/val_output13/epoch_{epoch}.npy'
    # gt_file_name = f'/home/cz_newacc/EE367/367_HDR/val_output_img/val_output13/gt_epoch_{epoch}.npy'
    np.save(file_name, output.cpu().numpy())
    np.save(gt_file_name, label.cpu().numpy())
    pbar.close()
    return sum(val_loss)/len(val_loss), sum(val_psnr)/len(val_loss)


def train_loop(train_dataloader, val_dataloader, num_epochs, learning_rate, momentum, device):
    # set up optimizer, loss, learning rate scheduler
    model = UNet(n_channels=9, n_classes=3) # RGB images
    model.to(device)
    for p in model.parameters():
        p.requires_grad = True
    optimizer = optim.Adam(model.parameters(), lr=learning_rate)
    criterion = nn.MSELoss()
    
    all_train_loss = []
    all_train_psnr = []
    all_val_loss = []
    all_val_psnr = []

    for epoch in range(num_epochs):
        train_loss, train_psnr, model = train_step(epoch, train_dataloader, model, optimizer, criterion)
        val_loss, val_psnr = validation_step(epoch, val_dataloader, model, criterion)
        all_train_loss.append(train_loss)
        all_train_psnr.append(train_psnr)
        # if epoch > 0 and val_psnr > max(all_val_psnr):
        #     torch.save(model.state_dict(), f'/home/cz_newacc/EE367/367_HDR/checkpoint/checkpoint13/epoch_{epoch}.pt')
        all_val_loss.append(val_loss)
        all_val_psnr.append(val_psnr)
    return all_train_loss, all_train_psnr, all_val_loss, all_val_psnr

if __name__ == '__main__':

    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    print(device)
    print(torch.cuda.is_available())
    
    # set parameters
    batchsize = 16
    num_epochs = 150
    learning_rate = 0.00005
    momentum = 0.9
 

    # prepare dataloader
    root_path = '/home/cz_newacc/EE367/367_HDR/data'
    dataloaders = prepare_dataloader(root_path, batchsize)

    # train loop
    all_train_loss, all_train_psnr, all_val_loss, all_val_psnr = train_loop(dataloaders['train'], dataloaders['val'],
                                                                            num_epochs, learning_rate, momentum, device)
    train_output = {'all_train_loss': all_train_loss,
                    'all_train_psnr': all_train_psnr,
                    'all_val_loss': all_val_loss,
                    'all_val_psnr': all_val_psnr}
    np.save('/home/cz_newacc/EE367/367_HDR/train_output/train_output13.npy', train_output)
    
    train_param = {'batchsize': batchsize,
                   'num_epochs': num_epochs,
                   'lr': learning_rate,
                   'momentum': momentum,
                  'optimizer': 'Adam',
                  'criterion': 'L2',
                  'regularizer':'lam=0.001, TV'}
    np.save('/home/cz_newacc/EE367/367_HDR/train_output/train_output_param13.npy', train_param)
                   