from curses import textpad
from pathlib import Path
from skimage.color import rgb2ycbcr, ycbcr2rgb, rgb2gray
import matplotlib.pyplot as plt
import skimage.io as io
import numpy as np
import glob
import re

def load_data(hdr_dir):
  ldr_imgs = []
  images = glob.glob(f'{hdr_dir}/*.png')
  images.sort(key=lambda x:[int(c) if c.isdigit() else c for c in re.split(r'(\d+)', x)])
  for i,image in enumerate(images):
    img = io.imread(image)
    img_norm = np.array(np.clip(img/255., a_min=0., a_max=1.))**2.2
    ldr_imgs.append(img_norm)
  return images, ldr_imgs


def luminance(ldr_imgs):
  ldr_imgs_ycbcr = []
  for i,img_norm in enumerate(ldr_imgs):
    img_ycbcr = rgb2ycbcr(img_norm)
    ldr_imgs_ycbcr.append(img_ycbcr)
    w = 1 / np.array([4., 2., 4.])
    hdr_y = np.clip(img_ycbcr[:,:,0]*w[i] if i==0 else hdr_y + img_ycbcr[:,:,0]*w[i], 16, 235)
    io.imsave(f'./ldr_y{i}.jpg', img_ycbcr[:,:,0], cmap='gray')
  io.imsave(f'./hdr_y.jpg', hdr_y, cmap='gray')
  return hdr_y


def chromaticity(ldr_imgs):
  w_list = []
  for img in ldr_imgs:
    img = img ** 2.2
    w = np.zeros_like(img)
    for xx in range(img.shape[0]):
      for yy in range(img.shape[1]):
        for cc in range(3):
          w[xx,yy,cc] = np.exp(-4 * (img[xx,yy,cc]-0.5)**2 / (0.5)**2)
    w_list.append(w)

  hdr = np.sum(np.array([np.multiply(w, (np.log(ldr_imgs[i]+np.finfo(np.float32).eps)-np.log(t[i]))) for i,w in enumerate(w_list)]), axis=0)
  scale = np.sum(w_list, axis=0)
  hdr = np.exp(hdr / scale)
  hdr *= 0.8371896 / np.mean(hdr)
  hdr_debevec = np.clip((1.*hdr)**(1./2.2), 0, 1)
  hdr = np.clip((1.*hdr)**(1./4.), 0, 1)
  hdr_cbcr = np.clip(rgb2ycbcr(hdr)[:,:,1:3], 16, 240)
  return hdr_debevec, hdr_cbcr


def calc_psnr(x, gt):
    out = 10 * np.log10(1 / ((x - gt)**2).mean().item())
    return out


if __name__ == '__main__':
  proj_dir = Path(__file__).parent
  hdr_dir = Path(proj_dir/"hdr_data")
  t = 1 / np.array([.125, 1, 64])
  images, ldr_imgs = load_data(hdr_dir=hdr_dir)
  # for i,ldr in enumerate(ldr_imgs):
  #   io.imsave(f'./ldr{i}.jpg', ldr[420:676, 252:508, :])
  hdr_y = luminance(ldr_imgs=ldr_imgs)
  hdr_debevec, hdr_cbcr = chromaticity(ldr_imgs=ldr_imgs, t_exp=t)
  io.imsave(f'./hdr_cb.jpg', hdr_cbcr[:,:,0], cmap='gray')
  io.imsave(f'./hdr_cr.jpg', hdr_cbcr[:,:,1], cmap='gray')
  hdr = np.clip(ycbcr2rgb(np.stack([hdr_y, hdr_cbcr[:,:,0], hdr_cbcr[:,:,1]], axis=2)), a_min=0., a_max=1.)
  # hdr_debevec, hdr = hdr_debevec[38:294, 30:286, :], hdr[38:294, 30:286, :] # 1
  # hdr_debevec, hdr = hdr_debevec[420:676, 252:508, :], hdr[420:676, 252:508, :] # 2

  io.imshow(hdr_debevec)
  io.imsave(f'./hdr_debevec.jpg', hdr_debevec)
  io.show()

  io.imshow(hdr)
  io.imsave(f'./hdr.jpg', hdr)
  io.show()
  
